#import "Controller.h"
#import "Stopwatch.h"


#define PEDANTIC	(1000*1000)
#define ACCURATE	(100*1000)
#define SLOPPY		(10*1000)

#define LOOPS	SLOPPY

// #define LOOPS   (1000*1000)   /* accurate, converts to us */

@implementation Controller

- (void) dealloc
{
   NSNotificationCenter   *center;

   center = [NSNotificationCenter defaultCenter];
   [center removeObserver:self];

   [super dealloc];
}


- (void) awakeFromNib
{
   NSNotificationCenter   *center;

   if( ! [self window])
   {
      [NSBundle loadNibNamed:@"Window"
                       owner:self];
      return;
   }
   
   center = [NSNotificationCenter defaultCenter];
   [center addObserver:self
              selector:@selector( textDidEndEnditing:)
                  name:NSControlTextDidEndEditingNotification
                object:nil];
}


extern void   alloc_loop( unsigned int pages, unsigned int loops);
extern void   touch_loop( unsigned int pages, unsigned int loops);
extern void   alloc_loop_mmap( unsigned int pages, unsigned int loops);
extern void   touch_loop_mmap( unsigned int pages, unsigned int loops);



- (void) textDidEndEnditing:(NSNotification *) notification
{
   double   alloc1;
   double   fault1;
   double   alloc2;
   double   fault2;
   double   pageextracost;
   double   timefault1;
   double   timefault2;
   double   allocationcost;
   double   faultzerocost;
   double   objectcost;
   double   zerofillcost;
   double   pages;
   double   t1;
   double   t2;
   double   t3;
   
   alloc1 = [[inputForm cellWithTag:0] doubleValue];
   fault1 = [[inputForm cellWithTag:1] doubleValue];
   alloc2 = [[inputForm cellWithTag:2] doubleValue];
   fault2 = [[inputForm cellWithTag:3] doubleValue];

   pageextracost = alloc2 - alloc1;

   timefault1 = fault1 - alloc1;
   timefault2 = fault2 - alloc2;

   objectcost = (2 * timefault1 - timefault2);

   allocationcost = (alloc1 - pageextracost) / 2;
   faultzerocost  = (timefault2 - timefault1);
   zerofillcost   = faultzerocost - (alloc1 / 2 / 4) - objectcost / 2;
   
   [[outputForm cellWithTag:0] setDoubleValue:allocationcost];
   [[outputForm cellWithTag:1] setDoubleValue:pageextracost];
   [[outputForm cellWithTag:2] setDoubleValue:objectcost];
   [[outputForm cellWithTag:3] setDoubleValue:faultzerocost];
   [[outputForm cellWithTag:4] setDoubleValue:zerofillcost];

   pages = [[calcForm cellWithTag:0] doubleValue];
   t1 = allocationcost + (pageextracost / 2) * pages;
   t2 = faultzerocost * pages + objectcost / 2;
   t3 = allocationcost + (pageextracost / 2) * pages + objectcost / 2;
   
   [[calcForm cellWithTag:1] setDoubleValue:t1];
   [[calcForm cellWithTag:2] setDoubleValue:t2];
   [[calcForm cellWithTag:3] setDoubleValue:pages * zerofillcost];
   [[calcForm cellWithTag:4] setDoubleValue:t3];
   [[calcForm cellWithTag:5] setDoubleValue:t1 + t2 + t3];
}


/* could keep lowest numbers ... */
- (double) testWithFunction:(void (*)( unsigned int, unsigned int)) f
                      pages:(unsigned int) pages
{
   Stopwatch       *watch;
   unsigned long   diff;
   double          value;
   
   watch = [[[Stopwatch alloc] init] autorelease];

   [watch start];
   (*f)( pages, LOOPS);
   [watch stop];
   diff = [watch elapsedMilliseconds];
   diff *= (1000*1000) / LOOPS;
   value = (double) diff / 1000.0;
   return( value);
}


- (void) setDoubleValue:(double) value
          inCellWithTag:(unsigned int) tag
                 ofForm:(NSForm *) form
              keepLower:(BOOL) keep
{
   NSCell  *cell;
   double  old;
   
   cell = [form cellWithTag:tag];
   if( keep)
   {
      old = [cell doubleValue];
      if( old < value)
         return;
   }
   [cell setDoubleValue:value];
   [form display];
}


- (void) benchmark:(id) sender
{
   unsigned int    i;
   BOOL            keep;
   double          value;
   BOOL            useMMap;
   
   keep    = [keepLowButton state] == NSOnState;
   useMMap = [mmapButton state] == NSOnState;
   
   if( ! flag)
      alloc_loop( 2, LOOPS);	/* warm up */

   for( i = 1; i <= 2; i++)
   {
      if( ! useMMap)
         value = [self testWithFunction:alloc_loop
                                  pages:i];
      else
         value = [self testWithFunction:alloc_loop_mmap
                                  pages:i];
         
      [self setDoubleValue:value
                      inCellWithTag:0 + i * 2  - 2
                    ofForm:inputForm
                 keepLower:keep];

      if(  ! useMMap)
         value = [self testWithFunction:touch_loop
                                  pages:i];
      else
         value = [self testWithFunction:touch_loop_mmap
                                  pages:i];
         
      [self setDoubleValue:value
             inCellWithTag:1 + i * 2  - 2
                    ofForm:inputForm
                 keepLower:keep];
   }
   [self textDidEndEnditing:nil];

   if( ! flag) /* good name :) */
   {
      flag = YES;
      [keepLowButton setState:NSOnState];
   }
}


- (void) benchmarkLoop:(id) sender
{
   double  old[ 4];
   int     i;
   BOOL    countdown;
   
   [self benchmark:sender];

again:      
   countdown = 2;
   
onemoretime:   
   for( i = 0; i < 4; i++)
      old[ i] = [[inputForm cellWithTag:i] doubleValue];
   
   [self benchmark:sender];

   for( i = 0; i < 4; i++)
      if( old[ i] != [[inputForm cellWithTag:i] doubleValue])
         goto again;

   if( ! countdown)
      return;
   --countdown;
   goto onemoretime;
}


- (void) anotherWindow:(id) sender
{
   // hackish
   Controller   *leak;

   leak = [[Controller alloc] init];
   [NSBundle loadNibNamed:@"Window"
                    owner:leak];
}


@end
