#import <Foundation/Foundation.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/mman.h>
#include <unistd.h>
#include <mach/mach.h>


#define USE_MMAP   1	/* chose yer API poison */


#if USE_MMAP

#define f_alloc	alloc_loop_mmap
#define f_touch touch_loop_mmap

#else

#define f_alloc	alloc_loop
#define f_touch touch_loop

#endif

/* Try this to see cost of zero page faulting 
 * in relationship to allocate and deallocate
 * START_PAGES=1 END_PAGES=3 LOOPS=1000*1000
 */
#define START_PAGES	1
#define END_PAGES       3

#define LOOPS   (1000*1000)

/* Try this to see cost of zero page faulting
 * in relationship to "other activtity" during
 * faulting. Some believe zerofill is cheap.
 * Check it out with Shark. Use a little fuzz,
 * set samples to 50 us and let run for 300000
 * samples or so
* START_PAGES=2048 END_PAGES=2048 LOOPS=1000
 */


void   alloc_loop( unsigned int pages, unsigned int loops)
{
   unsigned int   i;
   char           *block;

   for( i = 0; i < loops; i++)
   {
      if( vm_allocate( mach_task_self(), (vm_address_t *) &block, pages * vm_page_size, 1))
         abort();

      if( vm_deallocate( mach_task_self(), (vm_address_t) block, pages * vm_page_size))
         abort();
   }
}


void   alloc_loop_mmap( unsigned int pages, unsigned int loops)
{
   unsigned int   i;
   char           *block;

   for( i = 0; i < loops; i++)
   {
      block = mmap( NULL, pages * vm_page_size, PROT_READ|PROT_WRITE, MAP_ANON|MAP_PRIVATE, -1, 0);
      if( block == (char *) -1)
         abort();

      if( munmap( block, pages * vm_page_size) < 0)
         abort();
   }
}


void   touch_loop( unsigned int pages, unsigned int loops)
{
   unsigned int   i, j;
   char           *block;

   for( i = 0; i < loops; i++)
   {
      if( vm_allocate( mach_task_self(), (vm_address_t *) &block, pages * vm_page_size, 1))
         abort();

      for( j = 0; j < pages * vm_page_size; j += vm_page_size)
         block[ j] = 1;

      if( vm_deallocate( mach_task_self(), (vm_address_t) block, pages * vm_page_size))
         abort();
   }
}


void   touch_loop_mmap( unsigned int pages, unsigned int loops)
{
   unsigned int   i, j;
   char           *block;

   for( i = 0; i < loops; i++)
   {
      block = mmap( NULL, pages * vm_page_size, PROT_READ|PROT_WRITE, MAP_ANON|MAP_PRIVATE, -1, 0);
      if( block == (char *) -1)
         abort();

      for( j = 0; j < pages * vm_page_size; j += vm_page_size)
         block[ j] = 1;

      if( munmap( block, pages * vm_page_size) < 0)
         abort();
   }
}


#if 0
int  main()
{
   unsigned int   i;
   
   alloc_loop( END_PAGES, LOOPS);	/* warm up */

   for( i = START_PAGES; i <= END_PAGES; i++)
   {
      NSLog( @"Just alloc start: pages=%d\n", i);
      alloc_loop( i, LOOPS);
      NSLog( @"Just alloc stop: pages=%d\n", i);

      NSLog( @"Zero fill start: pages=%d\n", i);
      touch_loop( i, LOOPS);
      NSLog( @"Zero fill stop: pages=%d\n", i);
   }

   return( 0);
}
#endif   
